pub mod types;

use clap::Args;
use std::io::Write;
use tool_helper::{Error, Input};
use types::{LPC, MonoADPCMIterator, VAGADPCM, VAGHeader};

#[derive(Args)]
pub struct Arguments {
    name: Option<String>
}

pub fn convert(args: Arguments, output_file_path: &Option<std::path::PathBuf>, input: Input, output: &mut dyn Write) -> Result<(), Error> {
    let mut wav_file   = hound::WavReader::new(input)?;
    let     wav_header = wav_file.spec();

    validate(&wav_header)?;
    
    let     vagadpcm_samples = VAGHeader::expected_vagadpcm_samples(wav_file.len()) + 1;
    let mut lpc              = LPC::empty();

    tool_helper::raw::write_raw(output, &VAGHeader::create(vagadpcm_samples, wav_header.sample_rate, &get_name_for_file_header(args.name, output_file_path))?)?;
    for adpcm_sample in MonoADPCMIterator::create(wav_file.samples::<i16>()) {
        let (vagadpcm, new_lpc) = VAGADPCM::create(adpcm_sample?, lpc);
        tool_helper::raw::write_raw(output, &vagadpcm)?;   
        lpc = new_lpc;
    }
    tool_helper::raw::write_raw(output, &VAGADPCM::end())?;
    Ok(())
}

fn validate(wav_header: &hound::WavSpec) -> Result<(), Error> {
    if wav_header.sample_format != hound::SampleFormat::Int {
        return Err(Error::from_str("VAG: Only integer samples are supported as input."));
    }

    if wav_header.bits_per_sample != 16 {
        return Err(Error::from_str("VAG: Only 16bits samples are currently supported as input."));
    }

    if wav_header.channels != 1 {
        return Err(Error::from_str("VAG: Only mono samples are currently supported"));
    }

    Ok(())
}

fn get_name_for_file_header(name: Option<String>, output_file_path: &Option<std::path::PathBuf>) -> String {
    if let Some(name) = name {
        return name;
    }

    if let Some(output_file_path) = output_file_path {
        if let Some(file_name) = output_file_path.file_name() {
            let mut string = file_name.to_string_lossy().to_string();

            if let Some(idx) = string.rfind('.') {
                string.replace_range(idx.., "");
            }
            return string;
        }
    }

    return "<out>".to_owned();
}